use std::fs::File;
use std::fs::{OpenOptions};
use std::io::{Read};

// ## =----=----=----=----=

// Values for p_type

pub const PT_NULL: u32 = 0x0;
pub const PT_LOAD: u32 = 0x1;
pub const PT_DYNAMIC: u32 = 0x2;
pub const PT_INTERP: u32 = 0x3;
pub const PT_NOTE: u32 = 0x4;
pub const PT_SHLIB: u32 = 0x5;
pub const PT_PHDR: u32 = 0x6;
pub const PT_TLS: u32 = 0x7;
pub const PT_NUM: u32 = 0x8;
pub const PT_LOOS: u32 = 0x60000000;
pub const PT_GNU_EH_FRAME: u32 = 0x6474e550;
pub const PT_GNU_STACK: u32 = 0x6474e551;
pub const PT_GNU_RELRO: u32 = 0x6474e552;
pub const PT_LOSUNW: u32 = 0x6ffffffa;
pub const PT_SUNWBSS: u32 = 0x6ffffffa;
pub const PT_SUNWSTACK: u32 = 0x6ffffffb;
pub const PT_HISUNW: u32 = 0x6fffffff;
pub const PT_HIOS: u32 = 0x6fffffff;
pub const PT_LOPROC: u32 = 0x70000000;
pub const PT_HIPROC: u32 = 0x7fffffff;

// Values fot p_flags

pub const PF_X: u32 = (1 << 0);
pub const PF_W: u32 = (1 << 1);
pub const PF_R: u32 = (1 << 2);
pub const PF_MASKOS: u32 = 0x0ff00000;
pub const PF_MASKPROC: u32 = 0xf0000000;

// ## =----=----=----=----=

pub fn open_wrapper(filename: &str) -> Result<(Vec<u8>, File), ()> {

    let mut fd_w: File = OpenOptions::new()
        .read(true)
        .open(&filename)
        .expect("Unable to open file");
    
    let mut tbytes: Vec<u8> = Vec::<u8>::new();

    match fd_w.read_to_end(&mut tbytes) {
        Ok(_) => Ok((tbytes, fd_w)),
        Err(_) => Err(())
    }
}

// =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= x64 =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

/*
typedef struct
{
  unsigned char	e_ident[EI_NIDENT];	/* Magic number and other info */
  Elf64_Half	e_type;			/* Object file type */
  Elf64_Half	e_machine;		/* Architecture */
  Elf64_Word	e_version;		/* Object file version */
  Elf64_Addr	e_entry;		/* Entry point virtual address */
  Elf64_Off	e_phoff;		/* Program header table file offset */
  Elf64_Off	e_shoff;		/* Section header table file offset */
  Elf64_Word	e_flags;		/* Processor-specific flags */
  Elf64_Half	e_ehsize;		/* ELF header size in bytes */
  Elf64_Half	e_phentsize;		/* Program header table entry size */
  Elf64_Half	e_phnum;		/* Program header table entry count */
  Elf64_Half	e_shentsize;		/* Section header table entry size */
  Elf64_Half	e_shnum;		/* Section header table entry count */
  Elf64_Half	e_shstrndx;		/* Section header string table index */
} Elf64_Ehdr;
*/

type Elf64_Half = u16;
type Elf64_Word = u32;
type Elf64_Addr = u64;
type Elf64_Off = u64;
type Elf64_Xword = u64;

#[repr(C, packed)]
#[derive(Clone)]
pub struct Elf64_Ehdr {
    pub e_ident: [u8; 16],
    pub e_type: Elf64_Half,
    pub e_machine: Elf64_Half,
    pub e_version: Elf64_Word,
    pub e_entry: Elf64_Addr,
    pub e_phoff: Elf64_Off,
    pub e_shoff: Elf64_Off,
    pub e_flags: Elf64_Word,
    pub e_ehsize: Elf64_Half,
    pub e_phentsize: Elf64_Half,
    pub e_phnum: Elf64_Half,
    pub e_shentsize: Elf64_Half,
    pub e_shnum: Elf64_Half,
    pub e_shstrndx: Elf64_Half
}

impl Elf64_Ehdr {
    fn new() -> Elf64_Ehdr {
        Elf64_Ehdr {
            e_ident: [0; 16],
            e_type: 0,
            e_machine: 0,
            e_version: 0,
            e_entry: 0,
            e_phoff: 0,
            e_shoff: 0,
            e_flags: 0,
            e_ehsize: 0,
            e_phentsize: 0,
            e_phnum: 0,
            e_shentsize: 0,
            e_shnum: 0,
            e_shstrndx: 0
        }
    }
}

/*
typedef struct
{
  Elf64_Word	p_type;			/* Segment type */
  Elf64_Word	p_flags;		/* Segment flags */
  Elf64_Off	p_offset;		/* Segment file offset */
  Elf64_Addr	p_vaddr;		/* Segment virtual address */
  Elf64_Addr	p_paddr;		/* Segment physical address */
  Elf64_Xword	p_filesz;		/* Segment size in file */
  Elf64_Xword	p_memsz;		/* Segment size in memory */
  Elf64_Xword	p_align;		/* Segment alignment */
} Elf64_Phdr;
*/

#[repr(C, packed)]
#[derive(Clone, Debug)]
pub struct Elf64_Phdr {
    pub p_type: Elf64_Word,
    pub p_flags: Elf64_Word,
    pub p_offset: Elf64_Off,
    pub p_vaddr: Elf64_Addr,
    pub p_paddr: Elf64_Addr,
    pub p_filesz: Elf64_Xword,
    pub p_memsz: Elf64_Xword,
    pub p_align: Elf64_Xword
}

impl Elf64_Phdr {
    fn new() -> Elf64_Phdr {
      Elf64_Phdr {
          p_type: 0,
          p_flags: 0,
          p_offset: 0,
          p_vaddr: 0,
          p_filesz: 0,
          p_memsz: 0,
          p_align: 0,
          p_paddr: 0
      }  
    }
}

/*
typedef struct
{
  Elf64_Word	sh_name;		/* Section name (string tbl index) */
  Elf64_Word	sh_type;		/* Section type */
  Elf64_Xword	sh_flags;		/* Section flags */
  Elf64_Addr	sh_addr;		/* Section virtual addr at execution */
  Elf64_Off	sh_offset;		/* Section file offset */
  Elf64_Xword	sh_size;		/* Section size in bytes */
  Elf64_Word	sh_link;		/* Link to another section */
  Elf64_Word	sh_info;		/* Additional section information */
  Elf64_Xword	sh_addralign;		/* Section alignment */
  Elf64_Xword	sh_entsize;		/* Entry size if section holds table */
} Elf64_Shdr
*/

#[repr(C, packed)]
#[derive(Clone, Debug)]
pub struct Elf64_Shdr {
    pub sh_name: Elf64_Word,
    pub sh_type: Elf64_Word,
    pub sh_flags: Elf64_Xword,
    pub sh_addr: Elf64_Addr,
    pub sh_offset: Elf64_Off,
    pub sh_size: Elf64_Xword,
    pub sh_link: Elf64_Word,
  	pub sh_info: Elf64_Word,		/* Additional section information */
  	pub sh_addralign: Elf64_Xword,	/* Section alignment */
  	pub sh_entsize: Elf64_Xword		/* Entry size if section holds table */
}

impl Elf64_Shdr {
    fn new() -> Elf64_Shdr {
      Elf64_Shdr {
          sh_name: 0,
          sh_type: 0,
          sh_flags: 0,
          sh_offset: 0,
          sh_addr: 0,
          sh_size: 0,
          sh_link: 0,
          sh_info: 0,
          sh_addralign: 0,
          sh_entsize: 0
      }  
    }
}

pub fn parse_ehdr64(fbuf: &[u8]) -> Result<Elf64_Ehdr, ()> {
    let p: *const [u8; std::mem::size_of::<Elf64_Ehdr>()] = fbuf.as_ptr() as *const [u8; std::mem::size_of::<Elf64_Ehdr>()];

    Ok ( unsafe { std::mem::transmute(*p) } )
}

pub fn parse_phdr64(fbuf: &[u8], ehdr: &Elf64_Ehdr) -> Result<Vec<Elf64_Phdr>, ()> {
    let mut vec_phdr: Vec<Elf64_Phdr> = Vec::new();

    let pht: *const u8 = (fbuf.as_ptr() as u64 + ehdr.e_phoff as u64) as *const u8;

    for i in 0..ehdr.e_phnum {
        let p: *const [u8; std::mem::size_of::<Elf64_Phdr>()] = (pht as u64 + (i * ehdr.e_phentsize) as u64) as *const [u8; std::mem::size_of::<Elf64_Phdr>()];

        vec_phdr.push(unsafe { std::mem::transmute(*p) });
    }

    Ok(vec_phdr)
}

pub fn parse_shdr64(fbuf: &[u8], ehdr: &Elf64_Ehdr) -> Result<Vec<Elf64_Shdr>, ()> {
    let mut vec_shdr: Vec<Elf64_Shdr> = Vec::new();

    let sht: *const u8 = (fbuf.as_ptr() as u64 + ehdr.e_shoff as u64) as *const u8;

    for i in 0..ehdr.e_phnum {
        let p: *const [u8; std::mem::size_of::<Elf64_Shdr>()] = (sht as u64 + (i * ehdr.e_shentsize) as u64) as *const [u8; std::mem::size_of::<Elf64_Shdr>()];

        vec_shdr.push(unsafe { std::mem::transmute(*p) });
    }

    Ok(vec_shdr)
}

// =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= x64 =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

// =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= x32 =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

/*
typedef struct
{
  unsigned char	e_ident[EI_NIDENT];	/* Magic number and other info */
  Elf32_Half	e_type;			/* Object file type */
  Elf32_Half	e_machine;		/* Architecture */
  Elf32_Word	e_version;		/* Object file version */
  Elf32_Addr	e_entry;		/* Entry point virtual address */
  Elf32_Off	e_phoff;		/* Program header table file offset */
  Elf32_Off	e_shoff;		/* Section header table file offset */
  Elf32_Word	e_flags;		/* Processor-specific flags */
  Elf32_Half	e_ehsize;		/* ELF header size in bytes */
  Elf32_Half	e_phentsize;		/* Program header table entry size */
  Elf32_Half	e_phnum;		/* Program header table entry count */
  Elf32_Half	e_shentsize;		/* Section header table entry size */
  Elf32_Half	e_shnum;		/* Section header table entry count */
  Elf32_Half	e_shstrndx;		/* Section header string table index */
} Elf32_Ehdr;
*/

type Elf32_Half = u16;
type Elf32_Word = u32;
type Elf32_Addr = u32;
type Elf32_Off = u32;
type Elf32_Xword = u64;

#[repr(C, packed)]
#[derive(Clone)]
pub struct Elf32_Ehdr {
    pub e_ident: [u8; 16],
    pub e_type: Elf32_Half,
    pub e_machine: Elf32_Half,
    pub e_version: Elf32_Word,
    pub e_entry: Elf32_Addr,
    pub e_phoff: Elf32_Off,
    pub e_shoff: Elf32_Off,
    pub e_flags: Elf32_Word,
    pub e_ehsize: Elf32_Half,
    pub e_phentsize: Elf32_Half,
    pub e_phnum: Elf32_Half,
    pub e_shentsize: Elf32_Half,
    pub e_shnum: Elf32_Half,
    pub e_shstrndx: Elf32_Half
}

impl Elf32_Ehdr {
    fn new() -> Elf64_Ehdr {
        Elf64_Ehdr {
            e_ident: [0; 16],
            e_type: 0,
            e_machine: 0,
            e_version: 0,
            e_entry: 0,
            e_phoff: 0,
            e_shoff: 0,
            e_flags: 0,
            e_ehsize: 0,
            e_phentsize: 0,
            e_phnum: 0,
            e_shentsize: 0,
            e_shnum: 0,
            e_shstrndx: 0
        }
    }
}

/*
typedef struct
{
  Elf32_Word	p_type;			/* Segment type */
  Elf32_Off	p_offset;		/* Segment file offset */
  Elf32_Addr	p_vaddr;		/* Segment virtual address */
  Elf32_Addr	p_paddr;		/* Segment physical address */
  Elf32_Word	p_filesz;		/* Segment size in file */
  Elf32_Word	p_memsz;		/* Segment size in memory */
  Elf32_Word	p_flags;		/* Segment flags */
  Elf32_Word	p_align;		/* Segment alignment */
} Elf32_Phdr;
*/

#[repr(C, packed)]
#[derive(Clone, Debug)]
pub struct Elf32_Phdr {
    pub p_type: Elf32_Word,
    pub p_offset: Elf32_Off,
    pub p_vaddr: Elf32_Addr,
    pub p_paddr: Elf32_Addr,
    pub p_filesz: Elf32_Word,
    pub p_memsz: Elf32_Word,
    pub p_flags: Elf32_Word,
    pub p_align: Elf32_Word
}

impl Elf32_Phdr {
    fn new() -> Elf64_Phdr {
      Elf64_Phdr {
          p_type: 0,
          p_flags: 0,
          p_offset: 0,
          p_vaddr: 0,
          p_filesz: 0,
          p_memsz: 0,
          p_align: 0,
          p_paddr: 0
      }  
    }
}

/*
typedef struct
{
  Elf32_Word	sh_name;		/* Section name (string tbl index) */
  Elf32_Word	sh_type;		/* Section type */
  Elf32_Word	sh_flags;		/* Section flags */
  Elf32_Addr	sh_addr;		/* Section virtual addr at execution */
  Elf32_Off	sh_offset;		/* Section file offset */
  Elf32_Word	sh_size;		/* Section size in bytes */
  Elf32_Word	sh_link;		/* Link to another section */
  Elf32_Word	sh_info;		/* Additional section information */
  Elf32_Word	sh_addralign;		/* Section alignment */
  Elf32_Word	sh_entsize;		/* Entry size if section holds table */
} Elf32_Shdr;
*/

#[repr(C, packed)]
#[derive(Clone, Debug)]
pub struct Elf32_Shdr {
    pub sh_name: Elf32_Word,
    pub sh_type: Elf32_Word,
    pub sh_flags: Elf32_Word,
    pub sh_addr: Elf32_Addr,
    pub sh_offset: Elf32_Off,
    pub sh_size: Elf32_Word,
    pub sh_link: Elf32_Word,
  	pub sh_info: Elf32_Word,		/* Additional section information */
  	pub sh_addralign: Elf32_Word,	/* Section alignment */
  	pub sh_entsize: Elf32_Word		/* Entry size if section holds table */
}

impl Elf32_Shdr {
    fn new() -> Elf64_Shdr {
      Elf64_Shdr {
          sh_name: 0,
          sh_type: 0,
          sh_flags: 0,
          sh_offset: 0,
          sh_addr: 0,
          sh_size: 0,
          sh_link: 0,
          sh_info: 0,
          sh_addralign: 0,
          sh_entsize: 0
      }  
    }
}

pub fn parse_ehdr32(fbuf: &[u8]) -> Result<Elf32_Ehdr, ()> {
    let p: *const [u8; std::mem::size_of::<Elf32_Ehdr>()] = fbuf.as_ptr() as *const [u8; std::mem::size_of::<Elf32_Ehdr>()];

    Ok ( unsafe { std::mem::transmute(*p) } )
}

pub fn parse_phdr32(fbuf: &[u8], ehdr: &Elf32_Ehdr) -> Result<Vec<Elf32_Phdr>, ()> {
    let mut vec_phdr: Vec<Elf32_Phdr> = Vec::new();

    let pht: *const u8 = (fbuf.as_ptr() as u64 + ehdr.e_phoff as u64) as *const u8;

    for i in 0..ehdr.e_phnum {
        let p: *const [u8; std::mem::size_of::<Elf32_Phdr>()] = (pht as u64 + (i * ehdr.e_phentsize) as u64) as *const [u8; std::mem::size_of::<Elf32_Phdr>()];

        vec_phdr.push(unsafe { std::mem::transmute(*p) });
    }

    Ok(vec_phdr)
}

pub fn parse_shdr32(fbuf: &[u8], ehdr: &Elf32_Ehdr) -> Result<Vec<Elf32_Shdr>, ()> {
    let mut vec_shdr: Vec<Elf32_Shdr> = Vec::new();

    let sht: *const u8 = (fbuf.as_ptr() as u64 + ehdr.e_shoff as u64) as *const u8;

    for i in 0..ehdr.e_phnum {
        let p: *const [u8; std::mem::size_of::<Elf32_Shdr>()] = (sht as u64 + (i * ehdr.e_shentsize) as u64) as *const [u8; std::mem::size_of::<Elf32_Shdr>()];

        vec_shdr.push(unsafe { std::mem::transmute(*p) });
    }

    Ok(vec_shdr)
}

// =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-= x64 =-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
